@@ -42,6 +42,10 @@
 #include "terminal-prefs.h"
 #include "terminal-libgsystem.h"
 
+#ifdef ENABLE_SEARCH_PROVIDER
+#include "terminal-search-provider.h"
+#endif /* ENABLE_SEARCH_PROVIDER */
+
 #include <sys/wait.h>
 #include <errno.h>
 #include <string.h>
@@ -78,9 +82,15 @@ struct _TerminalApp
   GHashTable *encodings;
   gboolean encodings_locked;
 
+  GHashTable *screen_map;
+
   GSettings *global_settings;
   GSettings *desktop_interface_settings;
   GSettings *system_proxy_settings;
+
+#ifdef ENABLE_SEARCH_PROVIDER
+  TerminalSearchProvider *search_provider;
+#endif /* ENABLE_SEARCH_PROVIDER */
 };
 
 enum
@@ -209,7 +219,7 @@ terminal_app_encoding_list_notify_cb (GSettings   *settings,
                                       const char  *key,
                                       TerminalApp *app)
 {
-  gs_free char **encodings = NULL;
+  gs_strfreev char **encodings = NULL;
   int i;
   TerminalEncoding *encoding;
 
@@ -230,7 +240,7 @@ terminal_app_encoding_list_notify_cb (GSettings   *settings,
   if (terminal_encoding_is_valid (encoding))
     encoding->is_active = TRUE;
 
-  g_settings_get (settings, key, "^a&s", &encodings);
+  g_settings_get (settings, key, "^as", &encodings);
   for (i = 0; encodings[i] != NULL; ++i) {
       encoding = terminal_app_ensure_encoding (app, encodings[i]);
       if (!terminal_encoding_is_valid (encoding))
@@ -372,6 +382,8 @@ terminal_app_init (TerminalApp *app)
                     G_CALLBACK (terminal_app_encoding_list_notify_cb),
                     app);
 
+  app->screen_map = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+
   settings = g_settings_get_child (app->global_settings, "keybindings");
   terminal_accels_init (G_APPLICATION (app), settings);
 }
@@ -385,6 +397,7 @@ terminal_app_finalize (GObject *object)
                                         G_CALLBACK (terminal_app_encoding_list_notify_cb),
                                         app);
   g_hash_table_destroy (app->encodings);
+  g_hash_table_destroy (app->screen_map);
 
   g_object_unref (app->global_settings);
   g_object_unref (app->desktop_interface_settings);
@@ -411,6 +424,22 @@ terminal_app_dbus_register (GApplication    *application,
                                                                        error))
     return FALSE;
 
+#ifdef ENABLE_SEARCH_PROVIDER
+  if (g_settings_get_boolean (app->global_settings, TERMINAL_SETTING_SHELL_INTEGRATION_KEY)) {
+    gs_unref_object TerminalSearchProvider *search_provider;
+
+    search_provider = terminal_search_provider_new ();
+
+    if (!terminal_search_provider_dbus_register (app->search_provider,
+                                                 connection,
+                                                 TERMINAL_SEARCH_PROVIDER_PATH,
+                                                 error))
+      return FALSE;
+
+    gs_transfer_out_value (&app->search_provider, &search_provider);
+  }
+#endif /* ENABLE_SEARCH_PROVIDER */
+
   object = terminal_object_skeleton_new (TERMINAL_FACTORY_OBJECT_PATH);
   factory = terminal_factory_impl_new ();
   terminal_object_skeleton_set_factory (object, factory);
@@ -436,6 +465,14 @@ terminal_app_dbus_unregister (GApplication    *application,
     app->object_manager = NULL;
   }
 
+#ifdef ENABLE_SEARCH_PROVIDER
+  if (app->search_provider) {
+    terminal_search_provider_dbus_unregister (app->search_provider, connection, TERMINAL_SEARCH_PROVIDER_PATH);
+    g_object_unref (app->search_provider);
+    app->search_provider = NULL;
+  }
+#endif /* ENABLE_SEARCH_PROVIDER */
+
   G_APPLICATION_CLASS (terminal_app_parent_class)->dbus_unregister (application,
                                                                     connection,
                                                                     object_path);
@@ -519,6 +556,35 @@ terminal_app_new_terminal (TerminalApp     *app,
   return screen;
 }
 
+TerminalScreen *
+terminal_app_get_screen_by_uuid (TerminalApp *app,
+                                 const char  *uuid)
+{
+  return g_hash_table_lookup (app->screen_map, uuid);
+}
+
+void
+terminal_app_register_screen (TerminalApp *app,
+                              TerminalScreen *screen)
+{
+  const char *uuid;
+
+  uuid = terminal_screen_get_uuid (screen);
+  g_hash_table_insert (app->screen_map, g_strdup (uuid), screen);
+}
+
+void
+terminal_app_unregister_screen (TerminalApp *app,
+                                TerminalScreen *screen)
+{
+  gboolean found;
+  const char *uuid;
+
+  uuid = terminal_screen_get_uuid (screen);
+  found = g_hash_table_remove (app->screen_map, uuid);
+  g_assert (found == TRUE);
+}
+
 void
 terminal_app_edit_profile (TerminalApp     *app,
                            GSettings       *profile,
@@ -665,11 +731,11 @@ terminal_app_get_proxy_settings (TerminalApp *app)
 PangoFontDescription *
 terminal_app_get_system_font (TerminalApp *app)
 {
-  const char *font;
+  gs_free char *font = NULL;
 
   g_return_val_if_fail (TERMINAL_IS_APP (app), NULL);
 
-  g_settings_get (app->desktop_interface_settings, MONOSPACE_FONT_KEY_NAME, "&s", &font);
+  font = g_settings_get_string (app->desktop_interface_settings, MONOSPACE_FONT_KEY_NAME);
 
   return pango_font_description_from_string (font);
 }
